@@ -20,7 +20,6 @@ module Data.Iteratee.Iteratee (
   ,isStreamFinished
   -- ** Chunkwise Iteratees
   ,mapChunksM_
-  ,mapReduce
   ,getChunk
   ,getChunks
   -- ** Nested iteratee combinators
@@ -63,7 +62,6 @@ import Data.Iteratee.Base
 
 import Control.Exception
 import Control.Monad.Trans.Class
-import Control.Parallel
 import Data.Maybe
 import Data.Monoid
 import Data.Typeable
@@ -152,35 +150,6 @@ mapChunksM_ f = liftI step
     step s@(EOF _) = idone () s
 {-# INLINE mapChunksM_ #-}
 
--- | Perform a parallel map/reduce.  The `bufsize` parameter controls
--- the maximum number of chunks to read at one time.  A larger bufsize
--- allows for greater parallelism, but will require more memory.
--- 
--- Implementation of `sum`
--- 
--- > sum :: (Monad m, LL.ListLike s, Nullable s) => Iteratee s m Int64
--- > sum = getSum <$> mapReduce 4 (Sum . LL.sum)
-mapReduce ::
-  (Monad m, Nullable s, Monoid b)
-  => Int               -- ^ maximum number of chunks to read
-  -> (s -> b)          -- ^ map function
-  -> Iteratee s m b
-mapReduce bufsize f = liftI (step (0, []))
- where
-  step a@(!buf,acc) (Chunk xs)
-    | nullC xs = liftI (step a)
-    | buf >= bufsize =
-        let acc'  = mconcat acc
-            b'    = f xs
-        in b' `par` acc' `pseq` liftI (step (0,[b' `mappend` acc']))
-    | otherwise     =
-        let b' = f xs
-        in b' `par` liftI (step (succ buf,b':acc))
-  step (_,acc) s@(EOF Nothing) =
-    idone (mconcat acc) s
-  step acc       (EOF (Just err))  =
-    throwRecoverableErr err (step acc)
-
 -- | Get the current chunk from the stream.
 getChunk :: (Monad m, Nullable s, NullPoint s) => Iteratee s m s
 getChunk = liftI step
@@ -194,12 +163,12 @@ getChunk = liftI step
 
 -- | Get a list of all chunks from the stream.
 getChunks :: (Monad m, Nullable s) => Iteratee s m [s]
-getChunks = liftI (step [])
+getChunks = liftI (step id)
  where
   step acc (Chunk xs)
     | nullC xs    = liftI (step acc)
-    | otherwise   = liftI (step (xs:acc))
-  step acc stream = idone (reverse acc) stream
+    | otherwise   = liftI (step $ acc . (xs:))
+  step acc stream = idone (acc []) stream
 {-# INLINE getChunks #-}
 
 -- ---------------------------------------------------
