@@ -18,6 +18,7 @@ module Data.Iteratee.ListLike (
   ,dropWhile
   ,drop
   ,head
+  ,tryHead
   ,last
   ,heads
   ,peek
@@ -97,23 +98,13 @@ isFinished = liftI check
 -- |Read a stream to the end and return all of its elements as a list.
 -- This iteratee returns all data from the stream *strictly*.
 stream2list :: (Monad m, Nullable s, LL.ListLike s el) => Iteratee s m [el]
-stream2list = liftI (step [])
-  where
-    step acc (Chunk ls)
-      | nullC ls  = liftI (step acc)
-      | otherwise = liftI (step (acc ++ LL.toList ls))
-    step acc str  = idone acc str
+stream2list = liftM (concatMap LL.toList) getChunks
 {-# INLINE stream2list #-}
 
 -- |Read a stream to the end and return all of its elements as a stream.
 -- This iteratee returns all data from the stream *strictly*.
 stream2stream :: (Monad m, Nullable s, Monoid s) => Iteratee s m s
-stream2stream = icont (step mempty) Nothing
-  where
-    step acc (Chunk ls)
-      | nullC ls   = icont (step acc) Nothing
-      | otherwise  = icont (step (acc `mappend` ls)) Nothing
-    step acc str   = idone acc str
+stream2stream = liftM mconcat getChunks
 {-# INLINE stream2stream #-}
 
 
@@ -146,9 +137,12 @@ break cpred = icont (step mempty) Nothing
 
 
 -- |Attempt to read the next element of the stream and return it
--- Raise a (recoverable) error if the stream is terminated
+-- Raise a (recoverable) error if the stream is terminated.
 -- 
 -- The analogue of @List.head@
+-- 
+-- Because @head@ can raise an error, it shouldn't be used when constructing
+-- iteratees for @convStream@.  Use @tryHead@ instead.
 head :: (Monad m, LL.ListLike s el) => Iteratee s m el
 head = liftI step
   where
@@ -158,6 +152,17 @@ head = liftI step
   step stream      = icont step (Just (setEOF stream))
 {-# INLINE head #-}
 
+-- | Similar to @head@, except it returns @Nothing@ if the stream
+-- is terminated.
+tryHead :: (Monad m, LL.ListLike s el) => Iteratee s m (Maybe el)
+tryHead = liftI step
+  where
+  step (Chunk vec)
+    | LL.null vec  = liftI step
+    | otherwise    = idone (Just $ LL.head vec) (Chunk $ LL.tail vec)
+  step stream      = idone Nothing stream
+{-# INLINE tryHead #-}
+
 -- |Attempt to read the last element of the stream and return it
 -- Raise a (recoverable) error if the stream is terminated
 -- 
@@ -271,7 +276,7 @@ dropWhile p = liftI step
 length :: (Monad m, Num a, LL.ListLike s el) => Iteratee s m a
 length = liftI (step 0)
   where
-    step !i (Chunk xs) = liftI (step $! i + fromIntegral (LL.length xs))
+    step !i (Chunk xs) = liftI (step $ i + fromIntegral (LL.length xs))
     step !i stream     = idone i stream
 {-# INLINE length #-}
 
@@ -422,12 +427,7 @@ mapStream
      ,LooseMap s el el')
   => (el -> el')
   -> Enumeratee (s el) (s el') m a
-mapStream f = eneeCheckIfDone (liftI . step)
-  where
-    step k (Chunk xs)
-      | LL.null xs = liftI (step k)
-      | otherwise  = mapStream f $ k (Chunk $ lMap f xs)
-    step k s       = idone (liftI k) s
+mapStream f = mapChunks (lMap f)
 {-# SPECIALIZE mapStream :: Monad m => (el -> el') -> Enumeratee [el] [el'] m a #-}
 
 -- |Map the stream rigidly.
@@ -439,12 +439,7 @@ rigidMapStream
   :: (Monad m, LL.ListLike s el, NullPoint s)
   => (el -> el)
   -> Enumeratee s s m a
-rigidMapStream f = eneeCheckIfDone (liftI . step)
-  where
-    step k (Chunk xs)
-      | LL.null xs = liftI (step k)
-      | otherwise  = rigidMapStream f $ k (Chunk $ LL.rigidMap f xs)
-    step k s       = idone (liftI k) s
+rigidMapStream f = mapChunks (LL.rigidMap f)
 {-# SPECIALIZE rigidMapStream :: Monad m => (el -> el) -> Enumeratee [el] [el] m a #-}
 {-# SPECIALIZE rigidMapStream :: Monad m => (Word8 -> Word8) -> Enumeratee B.ByteString B.ByteString m a #-}
 
@@ -454,42 +449,54 @@ rigidMapStream f = eneeCheckIfDone (liftI . step)
 -- 
 -- The analogue of @List.filter@
 filter
-  :: (Monad m, Nullable s, LL.ListLike s el)
+  :: (Monad m, Functor m, Nullable s, LL.ListLike s el)
   => (el -> Bool)
   -> Enumeratee s s m a
-filter p = convStream f'
-  where
-    f' = icont step Nothing
-    step (Chunk xs)
-      | LL.null xs = f'
-      | otherwise  = idone (LL.filter p xs) mempty
-    step _ = f'
+filter p = convStream (LL.filter p <$> getChunk)
 {-# INLINE filter #-}
 
 -- |Creates an 'Enumeratee' in which elements from the stream are
--- grouped into @sz@-sized blocks.  The outer stream is completely
--- consumed and the final block may be smaller than \sz\.
+-- grouped into @sz@-sized blocks.  The final block may be smaller
+-- than \sz\.
 group
   :: (LL.ListLike s el, Monad m, Nullable s)
   => Int  -- ^ size of group
   -> Enumeratee s [s] m a
-group sz iinit = liftI $ go iinit LL.empty
-  where go icurr pfx (Chunk s) = case gsplit (pfx `LL.append` s) of 
-          (full, partial) | LL.null full -> liftI $ go icurr partial
-                          | otherwise    -> do inext <- lift $ enumPure1Chunk full icurr
-                                               liftI $ go inext partial
-        go icurr pfx (EOF mex) 
-          | LL.null pfx = lift . enumChunk (EOF mex) $ icurr
-          | otherwise = do inext <- lift $ enumPure1Chunk (LL.singleton pfx) icurr        
-                           lift . enumChunk (EOF mex) $ inext
-        gsplit ls = case LL.splitAt sz ls of
-          (g, rest) | LL.null rest -> if LL.length g == sz
-                                         then (LL.singleton g, LL.empty)
-                                         else (LL.empty, g)
-                    | otherwise -> let (grest, leftover) = gsplit rest
-                                       g' = g `LL.cons` grest
-                                   in g' `seq` (g', leftover)
-{-# INLINE group #-}
+group cksz iinit = liftI (step 0 id iinit)
+ where
+  -- there are two cases to consider for performance purposes:
+  --  1 - grouping lots of small chunks into bigger chunks
+  --  2 - breaking large chunks into smaller pieces
+  -- case 2 is easier, simply split a chunk into as many pieces as necessary
+  -- and pass them to the inner iteratee as one list.  @gsplit@ does this.
+  --
+  -- case 1 is a bit harder, need to hold onto each chunk and coalesce them
+  -- after enough have been received.  Currently using a difference list
+  -- for this, i.e ([s] -> [s])
+  --
+  -- not using eneeCheckIfDone because that loses final chunks at EOF
+  step sz pfxd icur (Chunk s)
+    | LL.null s               = liftI (step sz pfxd icur)
+    | LL.length s + sz < cksz = liftI (step (sz+LL.length s) (pfxd . (s:)) icur)
+    | otherwise               =
+        let (full, rest) = gsplit . mconcat $ pfxd [s]
+            pfxd'        = if LL.null rest then id else (rest:)
+        in  do
+              inext <- lift $ enumPure1Chunk full icur
+              liftI $ step (LL.length rest) pfxd' inext
+  step _ pfxd icur mErr = case pfxd [] of
+                         []   -> idone icur mErr
+                         rest -> do
+                           inext <- lift $ enumPure1Chunk [mconcat rest] icur
+                           idone inext mErr
+  gsplit ls = case LL.splitAt cksz ls of
+    (g, rest) | LL.null rest -> if LL.length g == cksz
+                                   then ([g], LL.empty)
+                                   else ([], g)
+              | otherwise -> let (grest, leftover) = gsplit rest
+                                 g' = g : grest
+                             in (g', leftover)
+
 
 -- | Creates an 'enumeratee' in which elements are grouped into
 -- contiguous blocks that are equal according to a predicate.
@@ -499,30 +506,74 @@ groupBy
   :: (LL.ListLike s el, Monad m, Nullable s)
   => (el -> el -> Bool)
   -> Enumeratee s [s] m a
-groupBy same iinit = liftI $ go iinit LL.empty
-    where go icurr pfx (Chunk s) = case gsplit (pfx `LL.append` s) of
-                                          (full, partial)
-                                              | LL.null full -> liftI $ go icurr partial
-                                              | otherwise -> do inext <- lift . enumPure1Chunk full $ icurr
-                                                                liftI $ go inext partial
-          go icurr pfx (EOF mex) 
-            | LL.null pfx = lift . enumChunk (EOF mex) $ icurr
-            | otherwise = do inext <- lift . enumPure1Chunk (LL.singleton pfx) $ icurr
-                             lift . enumChunk (EOF mex) $ inext
-          gsplit ll | LL.null ll = (LL.empty, LL.empty)
-                    | otherwise = let groups = llGroupBy same ll
-                                      full = LL.init groups
-                                      partial = LL.last groups
-                                  in full `seq` partial `seq` (full, partial)
-          llGroupBy eq l -- Copied from Data.ListLike, avoid spurious (Eq el) constraint
-              | LL.null l = LL.empty
-              | otherwise = LL.cons (LL.cons x ys) (llGroupBy eq zs)
-              where (ys, zs) = LL.span (eq x) xs
-                    x = LL.head l
-                    xs = LL.tail l
+groupBy same iinit = liftI $ go iinit (const True, id)
+  where 
+    -- As in group, need to handle grouping efficiently when we're fed
+    -- many small chunks.
+    -- 
+    -- Move the accumulation of groups by chunks into an accumulator
+    -- that runs through gsplit, which is pfx / partial here. When we
+    -- get a chunk, use gsplit to retrieve any full chunks and get the
+    -- carried accumulator.
+    -- 
+    -- At the end, "finish" the accumulator and handle the last chunk,
+    -- unless the stream was entirely empty and there is no
+    -- accumulator.
+    go icurr pfx (Chunk s) = case gsplit pfx s of
+      ([], partial)   -> liftI $ go icurr partial
+      (full, partial) -> do inext <- lift $ enumPure1Chunk full icurr
+                            liftI $ go inext partial
+    go icurr (_inpfx, pfxd) (EOF mex) = case pfxd [] of
+      [] -> lift . enumChunk (EOF mex) $ icurr
+      rest -> do inext <- lift . enumPure1Chunk [mconcat rest] $ icurr
+                 lift . enumChunk (EOF mex) $ inext
+    -- Here, gsplit carries an accumulator consisting of a predicate
+    -- "inpfx" that indicates whether a new element belongs in the
+    -- growing group, and a difference list to ultimately generate the
+    -- group.
+    --
+    -- The initial accumulator is a group that can accept anything and
+    -- is empty.
+    -- 
+    -- New chunks are split into groups. The cases are 
+
+    --   0. Trivially, empty chunk
+
+    --   1. One chunk, in the currently growing group: continue the
+    --   current prefix (and generate a new predicate, in case we had
+    --   the initial predicate
+    
+    --   2. One chunk, but not in the current group: finish the
+    --   current group and return a new accumulator for the
+    --   newly-started gorup
+    
+    --   3. Multiple chunks, the first of which completes the
+    --   currently growing group
+    
+    --   4. Multiple chunks, the first of which is a new group
+    --   separate from the currently-growing group
+    gsplit (inpfx, pfxd) curr = case llGroupBy same curr of
+      [] -> ([], (inpfx, pfxd))
+      [g0] | inpfx (LL.head g0) -> ([], (same $ LL.head g0, pfxd . (g0 :)))
+           | otherwise          -> ([mconcat $ pfxd []], (same $ LL.head g0, pfxd . (g0 :)))
+      (g0:grest@(_:_)) | inpfx (LL.head g0) -> let glast = Prelude.last grest
+                                                   gfirst = mconcat $ (pfxd . (g0 :)) []
+                                                   gdone = gfirst : Prelude.init grest
+                                               in ( gdone, (same (LL.head glast), (glast :)) )
+                       | otherwise -> let glast = Prelude.last grest
+                                          gfirst = mconcat $ pfxd []
+                                          gdone = gfirst : Prelude.init grest
+                                      in ( gdone, (same (LL.head glast), (glast :)) )
+    llGroupBy eq l -- Copied from Data.ListLike, avoid spurious (Eq el) constraint
+      | LL.null l = []
+      | otherwise = (LL.cons x ys):(llGroupBy eq zs)
+        where (ys, zs) = LL.span (eq x) xs
+              x = LL.head l
+              xs = LL.tail l
+
 {-# INLINE groupBy #-}
 
--- | Merge offers another way to nest iteratees: as a monad stack.
+-- | @merge@ offers another way to nest iteratees: as a monad stack.
 -- This allows for the possibility of interleaving data from multiple
 -- streams.
 -- 
@@ -532,7 +583,7 @@ groupBy same iinit = liftI $ go iinit LL.empty
 -- >
 -- > -- combine alternating lines from two sources
 -- > -- To see how this was derived, follow the types from
--- > -- 'ileaveStream logger' and work outwards.
+-- > -- 'ileaveLines logger' and work outwards.
 -- > run =<< enumFile 10 "file1" (joinI $ enumLinesBS $
 -- >           ( enumFile 10 "file2" . joinI . enumLinesBS $ joinI
 -- >                 (ileaveLines logger)) >>= run)
