@@ -21,9 +21,6 @@ import qualified Data.ListLike as LL
 import           Control.Monad as CM
 import           Control.Monad.Writer
 
-import Text.Printf (printf)
-import System.Environment (getArgs)
-
 instance Show (a -> b) where
   show _ = "<<function>>"
 
@@ -107,6 +104,11 @@ prop_head xs = P.length xs > 0 ==> runner1 (enumPure1Chunk xs Iter.head) == head
 prop_head2 xs = P.length xs > 0 ==> runner1 (enumPure1Chunk xs (Iter.head >> stream2list)) == tail xs
   where types = xs :: [Int]
 
+prop_tryhead xs = case xs of
+  [] -> runner1 (enumPure1Chunk xs tryHead) == Nothing
+  _  -> runner1 (enumPure1Chunk xs tryHead) == Just (P.head xs)
+  where types = xs :: [Int]
+
 prop_heads xs n = n > 0 ==>
  runner1 (enumSpecial xs n $ heads xs) == P.length xs
   where types = xs :: [Int]
@@ -316,11 +318,11 @@ prop_group xs n = n > 0 ==>
           where groupOne [] = Nothing
                 groupOne elts@(_:_) = Just . splitAt n $ elts
                            
-prop_groupBy xs m = m > 0 ==>
-                  runner2 (enumPure1Chunk xs $ Iter.groupBy pred stream2list)
-                  == runner1 (enumPure1Chunk (List.groupBy pred xs) stream2list)
+prop_groupBy xs = forAll (choose (2,5)) $ \m ->
+  let pred z1 z2 = (z1 `mod` m == z2 `mod` m)
+  in runner2 (enumPure1Chunk xs $ Iter.groupBy pred stream2list)
+       == runner1 (enumPure1Chunk (List.groupBy pred xs) stream2list)
   where types = xs :: [Int]
-        pred z1 z2 = (z1 `mod` m == z2 `mod` m)
 
 prop_mapM_ xs n = n > 0 ==>
  runWriter ((enumSpecial xs n (Iter.mapM_ f)) >>= run)
@@ -365,6 +367,7 @@ tests = [
     ,testProperty "break remainder" prop_break2
     ,testProperty "head" prop_head
     ,testProperty "head remainder" prop_head2
+    ,testProperty "tryhead" prop_tryhead
     ,testProperty "heads" prop_heads
     ,testProperty "null heads" prop_heads2
     ,testProperty "peek" prop_peek
